/* utilFileMem.c - Utility to manage persistent memory.
 * This package provides routines that substitute for malloc and free, and can allow the
 * user to save an enire memory structure on a file. This memory structure can be
 * instantly 'reloaded' (using UFM_READ as the mode). The file saving and loading
 * is done by way of memory mapping.
 *
*/

#include <sys/mman.h>
#include <fcntl.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <stdio.h>
#include "utilFileMem.h"


/* static variables for this package */
int fd = 0; /* the file descriptor of the memory block; 0 means not open */
int fullSize = 0; /* actual number of bytes in the file */
int maxSize = 0; /* size that the file could be grown to if necessary */
struct fileMem *fileMemHead = NULL; /* the address of the memory */
int mode = 0; /* UFM_READ, UFM_NEW or UFM_WRITE */
int memProt = 0; /* internal memory protection parm; depends on mode */
struct memBlock dummy; /* used in calculation of memBlockOverhead */
int memBlockOverhead; /* amound of non-data overhad in a block */


/* PRIMITIVE: initialize fixed size file mem partition */
int initFileMem(filename, requestedMode, numPages)
char *filename; /* the name of the memory file to use or create; if NULL use malloc */
int requestedMode; /* 1  create or overwrite; 2  write existing; 3 readonly */ 
int numPages; /* when creating this is the size of the file / UFM_BLOCK_SIZE */
{
    return initGrowableFileMem(filename, requestedMode, numPages, numPages);
}


/* PRIMITIVE: initialize growable memory - returns non-zero on failure */
int initGrowableFileMem(filename, requestedMode, numPages, maxPages)
char *filename; /* the name of the memory file to use or create; if NULL use malloc */
int requestedMode; /* 1  create or overwrite; 2  write existing; 3 readonly */
int numPages; /* when creating: the size of the file / UFM_BLOCK_SIZE (65536) */
int maxPages; /* when creating: largest possible size this DB can grow to */
  {
    int i, result, mapResult, openMode;
    char outbuffer[1024];
    char *cptr, *requiredAddress;

    /* 0. General initialization */
    memBlockOverhead = ((int)&((&dummy)->data) - (int)&((&dummy)->next));


    /* 1. Verify the subsystem is not already initialized */
    if(fileMemHead != NULL) return(UFM_ALREADY_INITIALIZED);


    /* 2. Determine file parameters */
    mode = requestedMode;
    if(mode == UFM_NEW) /* create */
      {
        /* check arguments */
        if (numPages < 1) return(UFM_MIN_SIZE_IS_ONE_PAGE);
        if (maxPages < numPages) return(UFM_MAX_LESS_THAN_MIN);

        fullSize = numPages * UFM_BLOCK_SIZE;
        maxSize = maxPages * UFM_BLOCK_SIZE;
        openMode = O_RDWR|O_CREAT|O_TRUNC;
        memProt = PROT_READ|PROT_WRITE;
      }
    else /* read or read/write */
      {
        struct stat statBuf;

        /* file must already exist */
        if(filename == NULL)
          {
            return(UFM_FILE_NOT_SPECIFIED);
          }
        if(stat(filename, &statBuf) != 0) return(UFM_FILE_NOT_FOUND);
        fullSize = statBuf.st_size;
        maxSize = fullSize; /* assume not growable for now */

        /* check that file size is a multiple of UFM_BLOCK_SIZE */
        if((fullSize % UFM_BLOCK_SIZE) != 0)
          return(UFM_FILE_NOT_BLOCK_MULTIPLE);

        if(mode == UFM_WRITE)
          {
            openMode = O_RDWR;
            memProt = PROT_READ|PROT_WRITE;
          }
        else /* mode must be UFM_READ */
          {
            if (mode != UFM_READ) return(UFM_UNKNOWN_MODE);
            openMode = O_RDONLY;
            memProt = PROT_READ;
          }
      }


    /* 3. Open the file */
    if(filename != NULL)
      {
        fd = open(filename, openMode);
        if (fd == -1)
          {
            fd = 0;
            return(UFM_COULD_NOT_OPEN_FILE);
          }

        /* ensure the file is writable by us and readable by others */
        if(mode == UFM_NEW)
          {
            fchmod(fd, S_IRUSR | S_IWUSR | S_IRGRP | S_IROTH);
          }
      }


    /* 4. Initialize file if creating */
    if(mode == UFM_NEW && filename != NULL)
      {
#ifndef SETBADDATA
	/* Create the file full of NULLs */
       result = ftruncate(fd, fullSize);
       if (result == -1)
         {
           close(fd);
           fd = 0;   
           return(UFM_COULD_NOT_CREATE_EMPTY);
         }

#else
        /* Initialize the buffer with text that will be recognized  as
           bad data during debugging operations */
        for(cptr = outbuffer; cptr < (outbuffer + 1024); cptr += 8)
          strcpy(cptr, "baddata");

        /* initialize the file in 1024 byte blocks*/
        for(i = 0; i<fullSize; i += 1024)
          {
            result = write(fd, outbuffer, 1024);
            if (result == -1)
              {
                close(fd);
                fd = 0;
                return(UFM_COULD_NOT_BLANK_EMPTY);
              }
          }
#endif
      }


  /* 5. Map the file to memory. */
  requiredAddress = MAX_MEM_LOCATION- maxSize;

  if(filename == NULL)
    {
      /* use ordinary malloc as the default */
      fileMemHead = (struct fileMem *)malloc(fullSize);
      if(fileMemHead == NULL) return(UFM_COULD_NOT_MALLOC);
    }
  else
    {

#ifndef ANYADDRESS
      mapResult = mapToMemory(requiredAddress);
#else
      /* the following code can be used to test cases when
         the above required address fails for one reason or another.
         The zero argument causes the system to choose the best location */ 
      mapResult = mapToMemory(0);
      printf("Address chosen was %x required address was %x\n", fileMemHead,
        requiredAddress);
#endif

      if(mapResult != 0) return(mapResult);

      /* If we are reusing an existing file, we might have actually mapped it in
         the wrong place, since there might have been space left over. The next step
         is to verify the location and remap if necessary */
      if(mode != UFM_NEW)
        {
          if(fileMemHead->correctFileMemHead != fileMemHead)
            {
              requiredAddress = (char *)fileMemHead->correctFileMemHead;
              munmap(fileMemHead, fullSize);
              mapResult = mapToMemory(requiredAddress);
              if(mapResult != 0) return(mapResult);

              maxSize = (int)(MAX_MEM_LOCATION - (int)fileMemHead);
            }
        }
    }

  
  /* 6. When creating, we must now initialize the whole memory as free space*/
  if(mode == UFM_NEW)
    {
      struct memBlock *firstFree;
      int offset;

      /* make sure the correctFileMemHead is set for future reference */
      fileMemHead->correctFileMemHead = fileMemHead;

      /* the first used block comes directly after the header aligned to 8 bytes*/
      offset = ROUNDUPTO(sizeof(struct fileMem),8);

      firstFree = (struct memBlock *)(((char *)fileMemHead) + offset);
      firstFree->next = NULL;
      firstFree->prev = NULL;
      firstFree->nextFree = NULL;
      firstFree->size = fullSize - offset;

      fileMemHead->firstFree = firstFree;
      fileMemHead->firstBlock = firstFree; /* everything is currently free */
      fileMemHead->lastBlock = firstFree;
    }
  
  return(0); 
}




/* PRIMITIVE: Get first block -- used when reading an existing file. The first block
   must be the root that points to all other data in the file */

char *getFirstBlockOfFileMem()
{
  if(fileMemHead == NULL) return(NULL);
  return((char *)&(fileMemHead->firstBlock)->data);
}



/* PRIMITIVE: Terminate */
terminateFileMem()
{
  if(fileMemHead == 0) return;
  if(fd == 0)
    free(fileMemHead);
  else
    {
      munmap(fileMemHead, fullSize);
      close(fd);
      fd = 0;
    }
  fileMemHead = 0;
}


/* PRIMITIVE: Allocate */
char *allocFileMem(size)
int size;
{
  return(allocFileMemRecommended(size, NULL));
}


/* ASSISTANT: Allocate with possible recommended spot that is known to be free
   because it was just freed - no checks are made as to the validity of the
   recommended spot, just its size */
char *allocFileMemRecommended(size, recommended)
int size;
struct memBlock *recommended;
  {
    struct memBlock *currentBlock, *prevBlock;
    int neededSize, attempt, growResult;

    /* 0. Initial checks */
    if(fileMemHead == 0) return((char *)malloc(size)); /* system not initialized */
    if(mode == UFM_READ) return(NULL);


    /* 1. Determine the amount of size needed, including overhead */
    neededSize = ROUNDUPTO((size + sizeof(struct memBlock) - sizeof(double)),8);


    /* 2. Find a free block of adequate size */
    if(recommended != NULL && recommended->size >= neededSize)
    {
       /* we are going to be allocate the recommended block which must also be the
          first on the free chain. This is used by realloc */
       currentBlock = recommended;
       prevBlock = NULL;
    }
    else for(attempt=1; attempt<3; attempt++)
    {
       /* in first attempt, try to find space; second attempt is after growing */
       for(prevBlock = NULL, currentBlock = fileMemHead->firstFree;
         currentBlock != NULL && currentBlock->size < neededSize;
         currentBlock = currentBlock->nextFree)
         {
           prevBlock = currentBlock;
         }

       if(currentBlock != NULL) break; /* we have found space */

       /* we have not found space; action depends on attempt number */
       /* on second attempt, report no memory */
       if(attempt == 2) return(NULL);

       /* on first attempt, try to grow memory file */
       growResult = growFileMem(neededSize);
       if(growResult != 0) return(NULL); /* no point trying again */
    }


    /* 3. If we have found more space than necessary, try splitting the free block
          into two free blocks. We will end up allocating the first of these, and
          while the second (newBlock) will be remain a smaller free block */
    if(currentBlock->size - neededSize > sizeof(struct memBlock))
      {
        struct memBlock * newBlock;
        int minSize;

        minSize = ROUNDUPTO(neededSize,8); /* to ensure new block aligns */
        newBlock = (struct memBlock *)(((char *)currentBlock) + minSize);
        newBlock->next = currentBlock->next;
        if(newBlock->next == NULL)
          {
            /* the new block will become the last block */
            fileMemHead->lastBlock = newBlock;
          }
        else
          {
             /* make the next block point back to this block */
             newBlock->next->prev = newBlock;
          }
        newBlock->prev = currentBlock;
        newBlock->nextFree = currentBlock->nextFree;
        newBlock->size = currentBlock->size - minSize;

        currentBlock->next = newBlock;
        currentBlock->nextFree = newBlock;
        currentBlock->size = minSize;
      }


    /* 4. Take currentblock out of the free list and return a pointer to its data */
    if(prevBlock == NULL)
      {
        /* this is the first free block */
        fileMemHead->firstFree = currentBlock->nextFree;
      }
    else
      {
        prevBlock->nextFree = currentBlock ->nextFree;
      }
    currentBlock->nextFree = UFM_ALLOCATED;

    return(char *)&currentBlock->data;
    
  }



/* PRIMITIVE: Reallocate. Returns the address on success, zero if nothing allocated */
char *reallocFileMem(address, size)
void *address;
int size;
  {
    struct memBlock *actualAddress;
    void *newaddress;
    char *och, *nch;
    int neededSize, i, canReallocate;

    /* if the size is zero, then free */
    if(size == 0)
      {
        freeFileMem(address);
        return(0);
      }

    /* if address is not set, then just malloc */
    if(address == NULL)
      {
        return(allocFileMem(size));
      }

    /* see if we can grow the current block - if so grow it */
    actualAddress = getActualAddress(address);
    if(actualAddress == 0) return(0);
    canReallocate = 0;
    neededSize = ROUNDUPTO((size + sizeof(struct memBlock) - sizeof(double)),8);
   
    if(actualAddress->size == neededSize)
    {
       /* no need to do anything */
       return(address);
    }

    if(actualAddress->size > neededSize)
    {
       /* request to shrink allocated block */
       canReallocate = 1;
    }
    else if(actualAddress->next->nextFree != UFM_ALLOCATED)
    {
       /* we know there is a free block next, let us see of we can expand the
          current block to use it (or possibly contact the block) */
       if((actualAddress->size + actualAddress->next->size) >= neededSize)
       {
          canReallocate = 1;
       }
    }

   if(canReallocate)
     {
       /* we now know that we can change the block - first free it*/
       freeFileMemPossReallocate(address, 1);
       /* then allocate it at the recommended spot */
       return(allocFileMemRecommended(size, actualAddress));
    }

    /* otherwise, just allocate a new block, and copy the data */
    newaddress = allocFileMem(size);
    if (newaddress == NULL) return NULL;
    for(och=(char *)address, nch=(char *)newaddress, i=0; i<size; och++, nch++, i++)
      {
        *nch = *och;
      } 
    freeFileMem(address);
    return(newaddress);
  }


/* PRIMITIVE: Free. */   
void freeFileMem(address)
char *address;
  {
     freeFileMemPossReallocate(address, 0);
  }

/* Free with possible preparation to reallocate */
void freeFileMemPossReallocate(address, aboutToReallocate)
char *address;
int aboutToReallocate;
  {
    struct memBlock *actualAddress;

    if(fileMemHead == 0)
      {
        free(address);
        return;
      }

    if(mode == UFM_READ) return;

    actualAddress = getActualAddress(address);
    if(actualAddress == 0) return;

    /* put the block back at the start of the free list */
    actualAddress->nextFree = fileMemHead->firstFree;
    fileMemHead->firstFree = actualAddress;

    /* now look to see if we can merge with a free block below ... */
    tryToMerge(actualAddress);
    if(!aboutToReallocate) tryToMerge(actualAddress->prev);

    return;
  }




/* PRIMITIVE: minimizeFileMem to the minimum size by dropping a trailing empty
   block. Returns 0 unless a system error occurs. If a system error occurs, the
   user should check to see whether the file is still mapped before using any
   of its memory. */
int minimizeFileMem()
{
  int minFullSize, proposedFullSize, remainingSize, mapResult, result, truncResult;
  struct memBlock *lastBlock, *currentBlock, *prevFreeBlock;

  result = 0;

  /* 1. Verify that the file is writable */
  if(mode == UFM_READ) return(UFM_INVALID_OP_ON_READ_ONLY);  


  /* 2. Compute how much space can be dropped */
  /* Proceed only if the last block is free */
  lastBlock = fileMemHead->lastBlock;
  if(lastBlock->nextFree == UFM_ALLOCATED) return(0);

  /* We can drop as much of the last block as possible, as long as the total
     file size remains a multiple of UFM_BLOCK_SIZE */
  minFullSize = fullSize - lastBlock->size;
  proposedFullSize = ROUNDUPTO(minFullSize, UFM_BLOCK_SIZE);
  remainingSize = proposedFullSize - minFullSize;

  /* Proceed only if we will actually be shortening the file! Actually
     we could use == in the following calcuation, because proposedFullSize
     could never have been calculated to be greater */
  if(proposedFullSize >= fullSize) return(0);


  /* 3. Adjust the blocks to delete the excess space, leaving nothing free,
        except perhaps a partially used block or internal fragments. */
  if(remainingSize > memBlockOverhead)
  {
    /* in the normal case, all we have to do is to adjust the size of the last block,
       because there is likely to remain a small last block, however we must
       not leave a block where the actual data size is not positive */
    lastBlock->size = remainingSize;
  }
  else
  {
    /* we must delete the last free block entirely, and increase the size of
       the preceding block */
    /* search for the previous block in free list */
    for(prevFreeBlock = NULL, currentBlock = fileMemHead->firstFree;
      currentBlock != lastBlock;
      currentBlock = currentBlock->nextFree)
        {
          prevFreeBlock = currentBlock;
        }
    /* make the previous free block skip this block */
    if(prevFreeBlock == NULL)
    {
      /* last block is the first */
      fileMemHead->firstFree = lastBlock->nextFree;
    }
    else
    {
      prevFreeBlock->nextFree = lastBlock->nextFree;
    }

    /* ensure the previous block in the main chain does not point to it */
    lastBlock->prev->next = NULL; 

    /* set the last block to be the previous one */
    fileMemHead->lastBlock = lastBlock->prev;

    /* increase the size of the previous block marginally */
    fileMemHead->lastBlock->size += remainingSize;
  }


  /* 4. Temporarily unmap */
  munmap(fileMemHead, fullSize);

  /* 5. Drop the space from the file, if any */
  truncResult = ftruncate(fd, proposedFullSize);
  if (truncResult == -1) result = UFM_COULD_NOT_MINIMIZE;

  /* 6. Remap */
  fullSize = proposedFullSize;
  mapResult = mapToMemory(fileMemHead);
  if (mapResult != 0) return(mapResult);

  return result;
}





/* ***************** INTERNAL UTILITIES ***************** */


/* utility to merge two adjacent blocks if they are free, to create
   a larger free block.
   If successful, block 2 is destroyed.
   This is necessary to prevent the worst kind of fragmentation. */
tryToMerge(block1)
struct memBlock *block1;
{
  struct memBlock *block2, *currentBlock, *prevBlock;

  /* only do it if we have two blocks... (a NULL as the argument would indicate
     we are before the beginning of the chain of blocks; a NULL as the
     next block would indicate we are past the end of the chain; in either
     case there is nothin to merge) */
  if(block1 == NULL) return;

  block2 = block1->next;
  if(block2 != NULL)
    {

      /* only do it if the blocks are actually free... */
      if((block1->nextFree != UFM_ALLOCATED) &&
        (block2->nextFree != UFM_ALLOCATED))
      {
        /* destroy block2 */
        block1->next = block2->next;
        if(block2->next != NULL)
        {
          /* make the subsequent block point back to the first block */
          block2->next->prev = block1;
        }
        else
        {
          /* we must be deleting the last block */
          fileMemHead->lastBlock = block1;
        }
        block1->size += block2->size;

        /* remove block2 from the free list */
        /* the search wouldn't be necessary if we kept back pointers, but
           that may involve excessive storage overhead */
        /* prevblock is the previous block in the free chain */
        if(block1->nextFree == block2)
          {
            /* shortcut the search -- this happens if we have a stack */
            prevBlock = block1;
          }
        else 
          {
            for(prevBlock = NULL, currentBlock = fileMemHead->firstFree;
              currentBlock != block2;
              currentBlock = currentBlock->nextFree)
              {
                prevBlock = currentBlock;
              }
          }
        
        if(prevBlock == NULL)
          {
            /* the block we are destroying is the very first in the free chain */
            fileMemHead->firstFree = block2->nextFree;
          }
        else /* prevBlock not NULL */
          {
            prevBlock->nextFree = block2->nextFree;
          }
      }
  }
}



/* growFileMem to about 110 % of its current size, rounded up to the nearest higher
   number of blocks. Called automatically when block allocation fails.
   minBytesNeeded is the size of the block that we are trying to install (including
   overhead) */
int growFileMem(minBytesNeeded)
int minBytesNeeded;
{
  int amountToGrow, growResult, proposedFullSize, mapResult, oldFullSize;
  struct memBlock *newBlock;

  /* 1. Verify that the file is writable */
  if(mode == UFM_READ) return(UFM_INVALID_OP_ON_READ_ONLY);

  /* 2. Verify that there is address space to allocate */
  if((maxSize - fullSize) < minBytesNeeded)
    return(UFM_COULD_NOT_GROW_SPACE_ENOUGH);

  /* 3. Temporarily unmap */
  munmap(fileMemHead, fullSize);
 
  /* 4. Determine how much we will grow */
  /* by default 10% more ... */
  amountToGrow = fullSize * 0.1;
  /* ... but at least the minimum requested */
  if(amountToGrow < minBytesNeeded) amountToGrow = minBytesNeeded;
  /* ... and no more than the maximum address space */
  if(amountToGrow > (maxSize - fullSize)) amountToGrow = maxSize - fullSize;
  /* ... all rounded up to the nearest block size */

  amountToGrow = ROUNDUPTO(amountToGrow, UFM_BLOCK_SIZE);

  oldFullSize = fullSize;
  proposedFullSize = fullSize + amountToGrow;

  /* 5. Grow the file (if we can!) */
  growResult = ftruncate(fd, proposedFullSize);
  if (growResult != -1) fullSize = proposedFullSize;

  /* 6. Remap */
  mapResult = mapToMemory(fileMemHead);
  if(mapResult != 0) return(mapResult);

  /* 7. Adjust the blocks to add this new one (if we were able to grow) */
  if(fullSize == proposedFullSize)
    {
      /* create the new block */
      newBlock = (struct memBlock *)(((char *)fileMemHead)+oldFullSize);
      newBlock->next = NULL;
      newBlock->prev = fileMemHead->lastBlock;
      newBlock->size = amountToGrow;

      /* hook up to original set of blocks (there must have been at least one) */
      fileMemHead->lastBlock->next = newBlock;
      fileMemHead->lastBlock = newBlock;

      /* hook up in free list as the first block */
      newBlock->nextFree = fileMemHead->firstFree;
      fileMemHead->firstFree = newBlock;

      /* merge this block with the previous free block, if there was any */
      tryToMerge(newBlock->prev);
    }

  /* 8. Return */
  if(fullSize != proposedFullSize)
    return(UFM_COULD_NOT_GROW_FILE_ENOUGH);
  return(0);

}


/* utility to map into memory; side effect: sets fileMemHead   
   returns 0 on success. On failure returns error code and closes    
   memory file */

int mapToMemory(address)
struct fileMem *address;
{
  int flags;

  if(address == (struct fileMem *)0)
    flags = MAP_SHARED;
  else
    flags = MAP_SHARED|MAP_FIXED;


  fileMemHead = (struct fileMem *)mmap(
    address, 
    fullSize,
    memProt,
    flags, fd, 0);

  if(fileMemHead == (struct fileMem *)-1)
    {
      /* failure to map memory */
      close(fd);
      fd = 0;
      fileMemHead = NULL;
      return(UFM_COULD_NOT_MAP);
    }
  return(0);
}

/* utility to determine the actual address of a block from a user's
   address. Returns zero if not an allocated block */
struct memBlock *getActualAddress(address)
char *address;
  {
    struct memBlock *actualAddress;

    if(fileMemHead == 0) return NULL;

    /* determine the actual address of the block */
    actualAddress = (struct memBlock *)(address - memBlockOverhead);

    /* verify the address is in the correct range of the file memory */
    if((int)actualAddress < (int)fileMemHead ||
      (int)actualAddress > (((int)fileMemHead) + fullSize))
      return NULL;
   
    /* verify the address is really allocated */
    if(actualAddress->nextFree != UFM_ALLOCATED)
      return NULL;

    return actualAddress;
  }



/* ***************** DEBUGGING ROUTINES ***************** */

void debugPrintMemOnFile(FILE*,char*);

/* DEBUG PRIMITIVE: debugPrintMem */
void debugPrintMem(address)
char *address;
{
  debugPrintMemOnFile(stdout, address);
}


/* DEBUG PRIMITIVE: debugPrintMemOnFile hex dump the memory of a particular block */
void debugPrintMemOnFile(file, address)
FILE *file;
char *address;
  {
    struct memBlock *actualAddress;
    int i, userSize,  userInts;
    int *ptr;

    /* determine the actual address of the block */
    actualAddress = getActualAddress(address);

    if(actualAddress == NULL)
    {
      fprintf(file, "*** Note: address %x is not in file memory or not \
allocated\n***Min possible address =%x Max=%x\n",
        address, fileMemHead->firstBlock, ((int)fileMemHead) + fullSize);
      return;
    }

    userSize = actualAddress->size- memBlockOverhead;
    userInts = userSize / 4;
    fprintf(file, "*** Addr: %x  Block:  %x  Usersize: %d\n", address,
      actualAddress, userSize);

    ptr=(int *)address;
    for(i=0; i<userInts; i++)
    {
      fprintf(file, "%8x ", *ptr);
      ptr++;
    }
    fprintf(file, "\n");
  }


/* DEBUG PRIMITIVE: debugDumpAll */
void debugDumpAll()
{
  debugPrintBlocksDetail(stdout, 2000, 1);
}

/* DEBUG PRIMITIVE: debugPrintBlocks */
debugPrintBlocks()
{
  debugPrintBlocksOnFile(stdout);
}


/* DEBUG PRIMITIVE: debugPrintBlocks on a specific file*/
debugPrintBlocksOnFile(file)
FILE *file;
{
  debugPrintBlocksDetail(file, 5, 0);
}


/* DEBUG PRIMITIVE: debugPrintBlocksDetail */
debugPrintBlocksDetail(file, maxBlocks, showDetail)
FILE *file;
int maxBlocks;
int showDetail;
{
  struct memBlock *currentBlock;
  int i = 0;
  int saidSkipping = 0;

  if(fileMemHead == 0)
    {
      fprintf(file,"No blocks: File memory NOT initialized\n");
      return;
    }

  fprintf(file, "  DebugPrintBlocks: First block = %x First free = %x Last block = %x\n",
    fileMemHead->firstBlock, fileMemHead->firstFree, fileMemHead->lastBlock);

  currentBlock = fileMemHead->firstBlock;
  while(currentBlock != NULL)
    {
      i++;
      if(i > maxBlocks && currentBlock->next != NULL && currentBlock->next->next != NULL)
      {
        if (!saidSkipping)
        {
           fprintf(file, " ... skipping to last 2 blocks ...\n");
           saidSkipping = 1;
        }
      }
      else
      {
        fprintf(file, " ... Block %d: add= %x  siz= %d  nxt= %x  prv= %x  nxfr= %x\n",
          i, currentBlock, currentBlock->size, currentBlock->next,
          currentBlock->prev, currentBlock->nextFree);
        if(showDetail != 0) debugPrintMemOnFile(
          file, ((char *)currentBlock) +memBlockOverhead);
      }
      currentBlock = currentBlock -> next;
    }
}



/* DEBUG PRIMITIVE: debugPrintFileMem */
void debugPrintFileMem()
{
  debugPrintFileMemOnFile(stdout);
}




/* DEBUG PRIMITIVE: debugPrintFileMem on a specific file */
debugPrintFileMemOnFile(file)
FILE *file;
{
  int numBlocks = 0;
  int numAlloc = 0;
  int largestAlloc = 0;
  int largestFree = 0;
  int totalAlloc = 0;
  int totalFree = 0;
  int numFree;
  struct memBlock *currentBlock;

  if(fileMemHead == 0)
    {
      fprintf(file,"File memory NOT initialized\n");
    }
  else
    {
      fprintf(file,"File memory initialized\n");
      if(fd == 0) fprintf(file," ...Using an malloc block\n");
      else fprintf(file," ...Using a mapped file\n");

      currentBlock = fileMemHead->firstBlock;
      while(currentBlock != NULL)
        {
          numBlocks ++;
          if(currentBlock->nextFree == UFM_ALLOCATED)
            {
              /* this block is allocated */
              numAlloc++;
              totalAlloc += currentBlock->size;
              if(currentBlock->size > largestAlloc) largestAlloc = currentBlock->size;
            }
          else
            {
              /* this block is free */
              totalFree += currentBlock->size;
              if(currentBlock->size > largestFree) largestFree = currentBlock->size;
            }
          currentBlock = currentBlock->next;
        }

      fprintf(file," ...full size = %d\n ...address = 0x%x\n ... number pf blocks = %d\n ...number allocated blocks = %d\n ...largest allocated block = %d\n ...largest free block =%d\n ...total allocated = %d\n ...total free =%d\n", fullSize, fileMemHead,








 numBlocks, numAlloc, largestAlloc, largestFree, totalAlloc, totalFree);

      fprintf(file," ...Sanity check 1: %d bytes overhead for header\n",
         fullSize - totalAlloc - totalFree); 

        /* run a sanity check on the free list */
        currentBlock = fileMemHead->firstFree;
        numFree = numBlocks - numAlloc;
        while(currentBlock != NULL)
          {
            totalFree -= currentBlock ->size;
            numFree --;
            currentBlock = currentBlock ->nextFree;
          }
        if(numFree != 0 || totalFree != 0)
          {
            fprintf(file," ...Sanity check 2 Failed. Free list has %d missing blocks, %d missing memory\n", numFree, totalFree);
          }
        else
          {
            fprintf(file," ...Sanity check 2 OK. Free list is fine\n");
          }

        debugPrintBlocksOnFile(file);
    }
}


/* To know if the file meme system is initialized and in which mode
 * (read/write) */
void fileMemState(int *initialized, int *openedInMode)
{
  *initialized  = (fileMemHead != NULL);  /* NULL means not initialized */
  *openedInMode = mode;          /* may be UFM_READ, UFM_WRITE, UFM_NEW */
}
